use std::collections::{HashMap, HashSet};
use std::env;
use std::path::{Path, PathBuf};
use std::process::exit;
use std::sync::Arc;

use anyhow::{bail, Result};

use crate::modules::{InitializationContext, ModuleContext, ModuleContextMqtt};
use crate::mqtt::OwnedTopicsService;
use crate::util::generate_alphanumeric_id;

mod config;
mod modules;
mod mqtt;
mod util;

#[derive(Debug)]
struct Paths {
    data_directory: Box<Path>,
    config: Box<Path>,
}

async fn get_paths_and_create_directories() -> Result<Paths> {
    let is_root: bool = users::get_current_uid() == 0;

    let paths = if is_root {
        Paths {
            config: PathBuf::from("/etc/hassliebe/config.toml").into_boxed_path(),
            data_directory: PathBuf::from("/var/lib/hassliebe").into_boxed_path(),
        }
    } else {
        let home = env::var_os("HOME").map(PathBuf::from);

        if let Some(home) = home {
            Paths {
                config: env::var_os("XDG_CONFIG_HOME")
                    .filter(|p| !p.is_empty())
                    .map(PathBuf::from)
                    .unwrap_or_else(|| home.clone().join(".config"))
                    .join("hassliebe/config.toml")
                    .into_boxed_path(),

                data_directory: env::var_os("XDG_DATA_HOME")
                    .filter(|p| !p.is_empty())
                    .map(PathBuf::from)
                    .unwrap_or_else(|| home.clone().join(".local/share"))
                    .join("hassliebe")
                    .into_boxed_path(),
            }
        } else {
            bail!("The HOME environment variable must be set when not running as root.")
        }
    };

    log::debug!("{:?}", &paths);

    if let Some(p) = paths.config.parent() {
        tokio::fs::create_dir_all(p).await?;
    }

    tokio::fs::create_dir_all(&paths.data_directory).await?;

    Ok(paths)
}

fn initialize_logger() {
    let mut builder = env_logger::builder();
    builder.parse_filters("warn,hassliebe=info");
    builder.parse_default_env();
    builder.init();
}

async fn load_machine_id(paths: &Paths) -> Result<String> {
    let overwrite_path = paths.data_directory.join("machine_id");
    let mut value = tokio::fs::read_to_string(&overwrite_path).await.unwrap_or("".to_owned()).trim().to_owned();

    if value.is_empty() {
        log::debug!("{} does not exist or is empty", overwrite_path.to_string_lossy());

        match tokio::fs::read_to_string("/etc/machine-id").await {
            Ok(id) => {
                value = id.trim().to_owned();
                log::debug!("Found machine ID in /etc/machine-id: {}", value);
            }
            Err(_) => {
                log::debug!("Could not read /etc/machine-id");

                value = generate_alphanumeric_id(32);
                log::info!("Generated new machine ID: {}", value);
                tokio::fs::write(overwrite_path, &value).await?;
            }
        }
    } else {
        log::debug!("Found machine ID in {}: {}", overwrite_path.to_string_lossy(), value);
    }

    Ok(value)
}

#[tokio::main]
async fn main() -> Result<()> {
    initialize_logger();

    if env::args().any(|a| a == "-v" || a == "--version" || a == "-h" || a == "--help") {
        println!("Hassliebe v{} is licensed under the Blue Oak Model License 1.0.0.", env!("CARGO_PKG_VERSION"));
        println!("See {} for more information.", env!("CARGO_PKG_HOMEPAGE"));
        exit(exitcode::OK);
    }

    let paths = get_paths_and_create_directories().await?;
    let machine_id = load_machine_id(&paths).await?;

    let Some(config) = config::load(&paths.config)? else {
        exit(exitcode::CONFIG);
    };

    let availability_topic = config.friendly_id.to_owned() + "/availability";
    let (mqtt_client, event_loop) = mqtt::create_client(&config, &machine_id, &availability_topic).await?;
    let discovery_device_object = mqtt::create_discovery_device_object(&config, &machine_id);

    let owned_topics_service = OwnedTopicsService::new(&paths.data_directory).await?;

    let mut initialization_context = InitializationContext {
        owned_mqtt_topics: HashSet::new(),
        message_handler_by_mqtt_topic: HashMap::new(),

        full: Arc::new(ModuleContext {
            config,
            machine_id,
            mqtt: ModuleContextMqtt {
                client: mqtt_client,
                availability_topic,
                discovery_device_object,
            },
            // dbus_session_connection: zbus::Connection::session().await.context("while connecting to the D-Bus session bus")?
        }),
    };

    modules::init_all(&mut initialization_context).await?;

    mqtt::start_communication(&initialization_context, event_loop, owned_topics_service).await
}
