use std::collections::{HashMap, HashSet};
use std::sync::Arc;

use anyhow::Result;
use json::JsonValue;
use rumqttc::{AsyncClient as MqttClient, ClientError, QoS};
use serde::Deserialize;
use validator::Validate;

mod buttons;
mod info;
mod notifications;

#[derive(Deserialize, Validate, Default)]
pub struct Config {
    #[validate]
    pub buttons: Option<buttons::Config>,

    #[validate]
    pub info: Option<info::Config>,

    #[validate]
    pub notifications: Option<notifications::Config>,
}

type MqttMessageHandler = dyn Fn(&str) -> Result<()>;

pub struct ModuleContextMqtt {
    pub discovery_device_object: JsonValue,
    pub availability_topic: String,
    pub client: MqttClient,
}

pub struct InitializationContext {
    // Owned topics are topics which a retained message was sent into.
    pub owned_mqtt_topics: HashSet<String>,
    pub message_handler_by_mqtt_topic: HashMap<String, Box<MqttMessageHandler>>,

    pub full: Arc<ModuleContext>,
}

impl InitializationContext {
    fn subscribe_mqtt_topic<F: Fn(&str) -> Result<()> + 'static>(&mut self, topic: impl Into<String>, handler: F) {
        self.message_handler_by_mqtt_topic.insert(topic.into(), Box::new(handler));
    }

    async fn publish_to_owned_mqtt_topic(&mut self, topic: &str, message: impl Into<String>) -> std::result::Result<(), ClientError> {
        let message = message.into();

        self.register_owned_mqtt_topic(topic);
        self.full.mqtt.client.publish(topic, QoS::AtLeastOnce, true, message).await
    }

    fn register_owned_mqtt_topic(&mut self, topic: &str) {
        self.owned_mqtt_topics.insert(topic.into());
    }

    fn get_full(&self) -> Arc<ModuleContext> {
        self.full.clone()
    }
}

pub struct ModuleContext {
    pub config: super::config::Config,
    pub mqtt: ModuleContextMqtt,
    pub machine_id: String,
    // pub dbus_session_connection: zbus::Connection,
}

impl ModuleContext {
    fn get_entity_id(&self, module_id: &str, sub_id: &str) -> String {
        format!("{}_{}_{}", self.config.friendly_id, module_id, sub_id)
    }
}

impl ModuleContextMqtt {
    fn get_homeassistant_topic(&self, component_type: &str, entity_id: &str, suffix: &str) -> String {
        format!("homeassistant/{}/{}/{}", component_type, entity_id, suffix)
    }
}

pub async fn init_all(context: &mut InitializationContext) -> Result<()> {
    buttons::init(context).await?;
    info::init(context).await?;
    notifications::init(context).await?;
    Ok(())
}
